@@ -12,11 +12,11 @@
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more client.
+ * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 /**
@@ -33,12 +33,16 @@
 #include <glib-object.h>
 #include <gudev/gudev.h>
 #include <libgnomeui/gnome-rr.h>
-#include <math.h>
+#include <cups/cups.h>
 
 #include "gcm-client.h"
+#include "gcm-device-xrandr.h"
+#include "gcm-device-udev.h"
+#include "gcm-device-cups.h"
+#include "gcm-device-sane.h"
+#include "gcm-device-virtual.h"
+#include "gcm-screen.h"
 #include "gcm-utils.h"
-#include "gcm-edid.h"
-#include "gcm-dmi.h"
 
 #include "egg-debug.h"
 
@@ -46,6 +50,8 @@ static void     gcm_client_finalize	(GObject     *object);
 
 #define GCM_CLIENT_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GCM_TYPE_CLIENT, GcmClientPrivate))
 
+static void gcm_client_xrandr_add (GcmClient *client, GnomeRROutput *output);
+
 /**
  * GcmClientPrivate:
  *
@@ -56,14 +62,18 @@ struct _GcmClientPrivate
 	gchar				*display_name;
 	GPtrArray			*array;
 	GUdevClient			*gudev_client;
-	GnomeRRScreen			*rr_screen;
-	GcmEdid				*edid;
-	GcmDmi				*dmi;
+	GcmScreen			*screen;
+	http_t				*http;
+	gboolean			 loading;
+	guint				 loading_refcount;
+	gboolean			 use_threads;
 };
 
 enum {
 	PROP_0,
 	PROP_DISPLAY_NAME,
+	PROP_LOADING,
+	PROP_USE_THREADS,
 	PROP_LAST
 };
 
@@ -77,6 +87,43 @@ static guint signals[SIGNAL_LAST] = { 0 };
 
 G_DEFINE_TYPE (GcmClient, gcm_client, G_TYPE_OBJECT)
 
+/**
+ * gcm_client_set_use_threads:
+ **/
+void
+gcm_client_set_use_threads (GcmClient *client, gboolean use_threads)
+{
+	client->priv->use_threads = use_threads;
+	g_object_notify (G_OBJECT (client), "use-threads");
+}
+
+/**
+ * gcm_client_set_loading:
+ **/
+static void
+gcm_client_set_loading (GcmClient *client, gboolean ret)
+{
+	client->priv->loading = ret;
+	g_object_notify (G_OBJECT (client), "loading");
+	egg_debug ("loading: %i", ret);
+}
+
+/**
+ * gcm_client_done_loading:
+ **/
+static void
+gcm_client_done_loading (GcmClient *client)
+{
+	static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
+
+	/* decrement refcount, with a lock */
+	g_static_mutex_lock (&mutex);
+	client->priv->loading_refcount--;
+	if (client->priv->loading_refcount == 0)
+		gcm_client_set_loading (client, FALSE);
+	g_static_mutex_unlock (&mutex);
+}
+
 /**
  * gcm_client_get_devices:
  *
@@ -182,79 +229,44 @@ out:
 }
 
 /**
- * gcm_client_gudev_add_type:
+ * gcm_client_gudev_add:
  **/
 static void
-gcm_client_gudev_add_type (GcmClient *client, GUdevDevice *udev_device, GcmDeviceType type)
+gcm_client_gudev_add (GcmClient *client, GUdevDevice *udev_device)
 {
-	gchar *title;
 	GcmDevice *device = NULL;
-	gchar *id;
+	GcmDevice *device_tmp = NULL;
 	gboolean ret;
+	const gchar *value;
 	GError *error = NULL;
-	const gchar *sysfs_path;
 	GcmClientPrivate *priv = client->priv;
-	gchar *serial = NULL;
-	gchar *manufacturer = NULL;
-	gchar *model = NULL;
 
-	/* add new device */
-	id = gcm_client_get_id_for_udev_device (udev_device);
-	title = g_strdup_printf ("%s - %s",
-				g_udev_device_get_property (udev_device, "ID_VENDOR"),
-				g_udev_device_get_property (udev_device, "ID_MODEL"));
+	/* matches our udev rules, which we can change without recompiling */
+	value = g_udev_device_get_property (udev_device, "GCM_DEVICE");
+	if (value == NULL)
+		goto out;
 
-	/* turn space delimiters into spaces */
-	g_strdelimit (title, "_", ' ');
+	/* create new device */
+	device = gcm_device_udev_new ();
+	ret = gcm_device_udev_set_from_device (device, udev_device, &error);
+	if (!ret) {
+		egg_debug ("failed to set for device: %s", error->message);
+		g_error_free (error);
+		goto out;
+	}
 
 	/* we might have a previous saved device with this ID, in which case nuke it */
-	device = gcm_client_get_device_by_id (client, id);
-	if (device != NULL) {
+	device_tmp = gcm_client_get_device_by_id (client, gcm_device_get_id (device));
+	if (device_tmp != NULL) {
+
 		/* remove from the array */
-		g_ptr_array_remove (client->priv->array, device);
+		g_ptr_array_remove (client->priv->array, device_tmp);
 
 		/* emit a signal */
-		egg_debug ("emit removed: %s", id);
-		g_signal_emit (client, signals[SIGNAL_REMOVED], 0, device);
-
-		/* unref our copy */
-		g_object_unref (device);
+		egg_debug ("emit removed: %s", gcm_device_get_id (device));
+		g_signal_emit (client, signals[SIGNAL_REMOVED], 0, device_tmp);
 	}
 
-	/* get sysfs path */
-	sysfs_path = g_udev_device_get_sysfs_path (udev_device);
-
-	/* create device */
-	device = gcm_device_new ();
-	manufacturer = g_strdup (g_udev_device_get_property (udev_device, "ID_VENDOR"));
-	model = g_strdup (g_udev_device_get_property (udev_device, "ID_MODEL"));
-	serial = g_strdup (g_udev_device_get_property (udev_device, "ID_SERIAL"));
-
-	/* get rid of underscores */
-	if (manufacturer != NULL) {
-		g_strdelimit (manufacturer, "_", ' ');
-		g_strchomp (manufacturer);
-	}
-	if (model != NULL) {
-		g_strdelimit (model, "_", ' ');
-		g_strchomp (model);
-	}
-	if (serial != NULL) {
-		g_strdelimit (serial, "_", ' ');
-		g_strchomp (serial);
-	}
-
-	g_object_set (device,
-		      "type", type,
-		      "id", id,
-		      "connected", TRUE,
-		      "serial", serial,
-		      "model", model,
-		      "manufacturer", manufacturer,
-		      "title", title,
-		      "native-device-sysfs", sysfs_path,
-		      NULL);
-
 	/* load the device */
 	ret = gcm_device_load (device, &error);
 	if (!ret) {
@@ -267,33 +279,13 @@ gcm_client_gudev_add_type (GcmClient *client, GUdevDevice *udev_device, GcmDevic
 	g_ptr_array_add (priv->array, g_object_ref (device));
 
 	/* signal the addition */
-	egg_debug ("emit: added %s to device list", id);
+	egg_debug ("emit: added %s to device list", gcm_device_get_id (device));
 	g_signal_emit (client, signals[SIGNAL_ADDED], 0, device);
 out:
 	if (device != NULL)
 		g_object_unref (device);
-	g_free (serial);
-	g_free (manufacturer);
-	g_free (model);
-	g_free (id);
-	g_free (title);
-}
-
-/**
- * gcm_client_gudev_add:
- **/
-static void
-gcm_client_gudev_add (GcmClient *client, GUdevDevice *udev_device)
-{
-	const gchar *value;
-
-	/* matches our udev rules, which we can change without recompiling */
-	value = g_udev_device_get_property (udev_device, "GCM_DEVICE");
-	if (value != NULL) {
-		value = g_udev_device_get_property (udev_device, "GCM_TYPE");
-		egg_debug ("found %s device: %s", value, g_udev_device_get_sysfs_path (udev_device));
-		gcm_client_gudev_add_type (client, udev_device, gcm_device_type_from_text (value));
-	}
+	if (device_tmp != NULL)
+		g_object_unref (device_tmp);
 }
 
 /**
@@ -309,10 +301,10 @@ gcm_client_uevent_cb (GUdevClient *gudev_client, const gchar *action, GUdevDevic
 }
 
 /**
- * gcm_client_add_connected_devices_usb:
+ * gcm_client_add_connected_devices_udev:
  **/
 static gboolean
-gcm_client_add_connected_devices_usb (GcmClient *client, GError **error)
+gcm_client_add_connected_devices_udev (GcmClient *client, GError **error)
 {
 	GList *devices;
 	GList *l;
@@ -335,143 +327,141 @@ gcm_client_add_connected_devices_usb (GcmClient *client, GError **error)
 
 	g_list_foreach (devices, (GFunc) g_object_unref, NULL);
 	g_list_free (devices);
+
+	/* inform the UI */
+	gcm_client_done_loading (client);
 	return TRUE;
 }
 
 /**
- * gcm_client_get_output_name:
- *
- * Return value: the output name, free with g_free().
+ * gcm_client_add_connected_devices_udev_thrd:
  **/
-static gchar *
-gcm_client_get_output_name (GcmClient *client, GnomeRROutput *output)
+static gpointer
+gcm_client_add_connected_devices_udev_thrd (GcmClient *client)
 {
-	const gchar *output_name;
-	gchar *name = NULL;
-	gchar *vendor = NULL;
-	GString *string;
-	gboolean ret;
-	guint width = 0;
-	guint height = 0;
-	GcmClientPrivate *priv = client->priv;
-
-	/* blank */
-	string = g_string_new ("");
-
-	/* if nothing connected then fall back to the connector name */
-	ret = gnome_rr_output_is_connected (output);
-	if (!ret)
-		goto out;
-
-	/* this is an internal panel, use the DMI data */
-	output_name = gnome_rr_output_get_name (output);
-	ret = gcm_utils_output_is_lcd_internal (output_name);
-	if (ret) {
-		/* find the machine details */
-		g_object_get (priv->dmi,
-			      "name", &name,
-			      "vendor", &vendor,
-			      NULL);
-
-		/* TRANSLATORS: this is the name of the internal panel */
-		output_name = _("Laptop LCD");
-	} else {
-		/* find the panel details */
-		g_object_get (priv->edid,
-			      "monitor-name", &name,
-			      "vendor-name", &vendor,
-			      NULL);
-	}
-
-	/* prepend vendor if available */
-	if (vendor != NULL && name != NULL)
-		g_string_append_printf (string, "%s - %s", vendor, name);
-	else if (name != NULL)
-		g_string_append (string, name);
-	else
-		g_string_append (string, output_name);
-
-out:
-	/* find the best option */
-	g_object_get (priv->edid,
-		      "width", &width,
-		      "height", &height,
-		      NULL);
-
-	/* append size if available */
-	if (width != 0 && height != 0) {
-		gfloat diag;
-
-		/* calculate the dialgonal using Pythagorean theorem */
-		diag = sqrtf ((powf (width,2)) + (powf (height, 2)));
-
-		/* print it in inches */
-		g_string_append_printf (string, " - %i\"", (guint) ((diag * 0.393700787f) + 0.5f));
-	}
-
-	g_free (name);
-	g_free (vendor);
-	return g_string_free (string, FALSE);
+	gcm_client_add_connected_devices_udev (client, NULL);
+	return NULL;
 }
 
 /**
- * gcm_client_get_id_for_xrandr_device:
+ * gcm_client_get_device_by_window_covered:
  **/
-static gchar *
-gcm_client_get_id_for_xrandr_device (GcmClient *client, GnomeRROutput *output)
+static gfloat
+gcm_client_get_device_by_window_covered (gint x, gint y, gint width, gint height,
+				         gint window_x, gint window_y, gint window_width, gint window_height)
 {
-	const gchar *output_name;
-	gchar *name = NULL;
-	gchar *vendor = NULL;
-	gchar *ascii = NULL;
-	gchar *serial = NULL;
-	GString *string;
-	gboolean ret;
-	GcmClientPrivate *priv = client->priv;
+	gfloat covered = 0.0f;
+	gint overlap_x;
+	gint overlap_y;
 
-	/* blank */
-	string = g_string_new ("xrandr");
-
-	/* if nothing connected then fall back to the connector name */
-	ret = gnome_rr_output_is_connected (output);
-	if (!ret)
+	/* to the right of the window */
+	if (window_x > x + width)
+		goto out;
+	if (window_y > y + height)
 		goto out;
 
-	/* find the best option */
-	g_object_get (priv->edid,
-		      "monitor-name", &name,
-		      "vendor-name", &vendor,
-		      "ascii-string", &ascii,
-		      "serial-number", &serial,
-		      NULL);
+	/* to the left of the window */
+	if (window_x + window_width < x)
+		goto out;
+	if (window_y + window_height < y)
+		goto out;
 
-	/* append data if available */
-	if (vendor != NULL)
-		g_string_append_printf (string, "_%s", vendor);
-	if (name != NULL)
-		g_string_append_printf (string, "_%s", name);
-	if (ascii != NULL)
-		g_string_append_printf (string, "_%s", ascii);
-	if (serial != NULL)
-		g_string_append_printf (string, "_%s", serial);
+	/* get the overlaps */
+	overlap_x = MIN((window_x + window_width - x), width) - MAX(window_x - x, 0);
+	overlap_y = MIN((window_y + window_height - y), height) - MAX(window_y - y, 0);
+
+	/* not in this window */
+	if (overlap_x <= 0)
+		goto out;
+	if (overlap_y <= 0)
+		goto out;
+
+	/* get the coverage */
+	covered = (gfloat) (overlap_x * overlap_y) / (gfloat) (window_width * window_height);
+	egg_debug ("overlap_x=%i,overlap_y=%i,covered=%f", overlap_x, overlap_y, covered);
 out:
-	/* fallback to the output name */
-	if (string->len == 6) {
-		output_name = gnome_rr_output_get_name (output);
-		ret = gcm_utils_output_is_lcd_internal (output_name);
-		if (ret)
-			output_name = "LVDS";
-		g_string_append (string, output_name);
+	return covered;
+}
+
+/**
+ * gcm_client_get_device_by_window:
+ **/
+GcmDevice *
+gcm_client_get_device_by_window (GcmClient *client, GdkWindow *window)
+{
+	gfloat covered;
+	gfloat covered_max = 0.0f;
+	gint window_width, window_height;
+	gint window_x, window_y;
+	gint x, y;
+	guint i;
+	guint len = 0;
+	guint width, height;
+	GnomeRRMode *mode;
+	GnomeRROutput *output_best = NULL;
+	GnomeRROutput **outputs;
+	GcmDevice *device = NULL;
+	gchar *id = NULL;
+
+	/* get the window parameters, in root co-ordinates */
+	gdk_window_get_origin (window, &window_x, &window_y);
+	gdk_drawable_get_size (GDK_DRAWABLE(window), &window_width, &window_height);
+
+	/* get list of updates */
+	outputs = gcm_screen_get_outputs (client->priv->screen, NULL);
+	if (outputs == NULL)
+		goto out;
+
+	/* find length */
+	for (i=0; outputs[i] != NULL; i++)
+		len++;
+
+	/* go through each option */
+	for (i=0; i<len; i++) {
+
+		/* not interesting */
+		if (!gnome_rr_output_is_connected (outputs[i]))
+			continue;
+
+		/* get details about the output */
+		gnome_rr_output_get_position (outputs[i], &x, &y);
+		mode = gnome_rr_output_get_current_mode (outputs[i]);
+		width = gnome_rr_mode_get_width (mode);
+		height = gnome_rr_mode_get_height (mode);
+		egg_debug ("%s: %ix%i -> %ix%i (%ix%i -> %ix%i)", gnome_rr_output_get_name (outputs[i]),
+			   x, y, x+width, y+height,
+			   window_x, window_y, window_x+window_width, window_y+window_height);
+
+		/* get the fraction of how much the window is covered */
+		covered = gcm_client_get_device_by_window_covered (x, y, width, height,
+								   window_x, window_y, window_width, window_height);
+
+		/* keep a running total of which one is best */
+		if (covered > 0.01f && covered > covered_max) {
+			output_best = outputs[i];
+
+			/* optimize */
+			if (covered > 0.99) {
+				egg_debug ("all in one window, no need to search other windows");
+				goto out;
+			}
+
+			/* keep looking */
+			covered_max = covered;
+			egg_debug ("personal best of %f for %s", covered, gnome_rr_output_get_name (output_best));
+		}
 	}
-
-	/* replace unsafe chars */
-	gcm_utils_alphanum_lcase (string->str);
-
-	g_free (name);
-	g_free (vendor);
-	g_free (ascii);
-	g_free (serial);
-	return g_string_free (string, FALSE);
+out:
+	/* if we found an output, get the device */
+	if (output_best != NULL) {
+		GcmDevice *device_tmp;
+		device_tmp = gcm_device_xrandr_new ();
+		gcm_device_xrandr_set_from_output (device_tmp, output_best, NULL);
+		device = gcm_client_get_device_by_id (client, gcm_device_get_id (device_tmp));
+		g_free (device_tmp);
+	}
+	g_free (id);
+	return device;
 }
 
 /**
@@ -480,17 +470,11 @@ out:
 static void
 gcm_client_xrandr_add (GcmClient *client, GnomeRROutput *output)
 {
-	gchar *title = NULL;
-	gchar *id = NULL;
 	gboolean ret;
-	GcmDevice *device = NULL;
-	GError *error = NULL;
-	const gchar *output_name;
-	gchar *serial = NULL;
-	gchar *manufacturer = NULL;
-	gchar *model = NULL;
-	const guint8 *data;
 	gboolean connected;
+	GError *error = NULL;
+	GcmDevice *device = NULL;
+	GcmDevice *device_tmp = NULL;
 	GcmClientPrivate *priv = client->priv;
 
 	/* if nothing connected then ignore */
@@ -500,29 +484,21 @@ gcm_client_xrandr_add (GcmClient *client, GnomeRROutput *output)
 		goto out;
 	}
 
-	/* parse the EDID to get a crtc-specific name, not an output specific name */
-	data = gnome_rr_output_get_edid_data (output);
-	if (data != NULL) {
-		ret = gcm_edid_parse (priv->edid, data, NULL);
-		if (!ret) {
-			egg_warning ("failed to parse edid");
-			goto out;
-		}
-	} else {
-		/* reset, as not available */
-		gcm_edid_reset (priv->edid);
+	/* create new device */
+	device = gcm_device_xrandr_new ();
+	ret = gcm_device_xrandr_set_from_output (device, output, &error);
+	if (!ret) {
+		egg_debug ("failed to set for output: %s", error->message);
+		g_error_free (error);
+		goto out;
 	}
 
-	/* get details */
-	id = gcm_client_get_id_for_xrandr_device (client, output);
-	egg_debug ("asking to add %s", id);
-
 	/* we might have a previous saved device with this ID, in which case nuke it */
-	device = gcm_client_get_device_by_id (client, id);
-	if (device != NULL) {
+	device_tmp = gcm_client_get_device_by_id (client, gcm_device_get_id (device));
+	if (device_tmp != NULL) {
 
 		/* old device not connected */
-		g_object_get (device,
+		g_object_get (device_tmp,
 			      "connected", &connected,
 			      NULL);
 
@@ -533,46 +509,13 @@ gcm_client_xrandr_add (GcmClient *client, GnomeRROutput *output)
 		}
 
 		/* remove from the array */
-		g_ptr_array_remove (client->priv->array, device);
+		g_ptr_array_remove (client->priv->array, device_tmp);
 
 		/* emit a signal */
-		egg_debug ("emit removed: %s (unconnected device that has just been connected)", id);
-		g_signal_emit (client, signals[SIGNAL_REMOVED], 0, device);
-
-		/* unref our copy */
-		g_object_unref (device);
+		egg_debug ("emit removed: %s (unconnected device that has just been connected)", gcm_device_get_id (device_tmp));
+		g_signal_emit (client, signals[SIGNAL_REMOVED], 0, device_tmp);
 	}
 
-	/* get data about the display */
-	g_object_get (priv->edid,
-		      "monitor-name", &model,
-		      "vendor-name", &manufacturer,
-		      "serial-number", &serial,
-		      NULL);
-
-	/* refine data if it's missing */
-	output_name = gnome_rr_output_get_name (output);
-	ret = gcm_utils_output_is_lcd_internal (output_name);
-	if (ret && model == NULL) {
-		g_object_get (priv->dmi,
-			      "version", &model,
-			      NULL);
-	}
-
-	/* add new device */
-	device = gcm_device_new ();
-	title = gcm_client_get_output_name (client, output);
-	g_object_set (device,
-		      "type", GCM_DEVICE_TYPE_DISPLAY,
-		      "id", id,
-		      "connected", TRUE,
-		      "serial", serial,
-		      "model", model,
-		      "manufacturer", manufacturer,
-		      "title", title,
-		      "native-device-xrandr", output_name,
-		      NULL);
-
 	/* load the device */
 	ret = gcm_device_load (device, &error);
 	if (!ret) {
@@ -585,33 +528,13 @@ gcm_client_xrandr_add (GcmClient *client, GnomeRROutput *output)
 	g_ptr_array_add (priv->array, g_object_ref (device));
 
 	/* signal the addition */
-	egg_debug ("emit: added %s to device list", id);
+	egg_debug ("emit: added %s to device list", gcm_device_get_id (device));
 	g_signal_emit (client, signals[SIGNAL_ADDED], 0, device);
 out:
 	if (device != NULL)
 		g_object_unref (device);
-	g_free (id);
-	g_free (serial);
-	g_free (manufacturer);
-	g_free (model);
-	g_free (title);
-}
-
-/**
- * gcm_client_randr_event_cb:
- **/
-static void
-gcm_client_randr_event_cb (GnomeRRScreen *screen, GcmClient *client)
-{
-	GnomeRROutput **outputs;
-	guint i;
-
-	egg_debug ("screens may have changed");
-
-	/* replug devices */
-	outputs = gnome_rr_screen_list_outputs (client->priv->rr_screen);
-	for (i=0; outputs[i] != NULL; i++)
-		gcm_client_xrandr_add (client, outputs[i]);
+	if (device_tmp != NULL)
+		g_object_unref (device_tmp);
 }
 
 /**
@@ -624,12 +547,87 @@ gcm_client_add_connected_devices_xrandr (GcmClient *client, GError **error)
 	guint i;
 	GcmClientPrivate *priv = client->priv;
 
-	outputs = gnome_rr_screen_list_outputs (priv->rr_screen);
+	outputs = gcm_screen_get_outputs (priv->screen, error);
+	if (outputs == NULL)
+		return FALSE;
 	for (i=0; outputs[i] != NULL; i++)
 		gcm_client_xrandr_add (client, outputs[i]);
 	return TRUE;
 }
 
+/**
+ * gcm_client_cups_add:
+ **/
+static void
+gcm_client_cups_add (GcmClient *client, cups_dest_t dest)
+{
+	gboolean ret;
+	GError *error = NULL;
+	GcmDevice *device = NULL;
+	GcmClientPrivate *priv = client->priv;
+
+	/* create new device */
+	device = gcm_device_cups_new ();
+	ret = gcm_device_cups_set_from_dest (device, priv->http, dest, &error);
+	if (!ret) {
+		egg_debug ("failed to set for output: %s", error->message);
+		g_error_free (error);
+		goto out;
+	}
+
+	/* load the device */
+	ret = gcm_device_load (device, &error);
+	if (!ret) {
+		egg_warning ("failed to load: %s", error->message);
+		g_error_free (error);
+		goto out;
+	}
+
+	/* add to the array */
+	g_ptr_array_add (priv->array, g_object_ref (device));
+
+	/* signal the addition */
+	egg_debug ("emit: added %s to device list", gcm_device_get_id (device));
+	g_signal_emit (client, signals[SIGNAL_ADDED], 0, device);
+out:
+	if (device != NULL)
+		g_object_unref (device);
+}
+
+/**
+ * gcm_client_add_connected_devices_cups:
+ **/
+static gboolean
+gcm_client_add_connected_devices_cups (GcmClient *client, GError **error)
+{
+	gint num_dests;
+	cups_dest_t *dests;
+	gint i;
+	GcmClientPrivate *priv = client->priv;
+
+	num_dests = cupsGetDests2 (priv->http, &dests);
+	egg_debug ("got %i printers", num_dests);
+
+	/* get printers on the local server */
+	for (i = 0; i < num_dests; i++)
+		gcm_client_cups_add (client, dests[i]);
+	cupsFreeDests (num_dests, dests);
+
+	/* inform the UI */
+	gcm_client_done_loading (client);
+	return TRUE;
+}
+
+/**
+ * gcm_client_add_connected_devices_cups_thrd:
+ **/
+static gpointer
+gcm_client_add_connected_devices_cups_thrd (GcmClient *client)
+{
+	gcm_client_add_connected_devices_cups (client, NULL);
+	return NULL;
+}
+
 /**
  * gcm_client_add_unconnected_device:
  **/
@@ -638,9 +636,12 @@ gcm_client_add_unconnected_device (GcmClient *client, GKeyFile *keyfile, const g
 {
 	gchar *title;
 	gchar *type_text = NULL;
-	GcmDeviceType type;
+	gchar *colorspace_text = NULL;
+	GcmColorspaceEnum colorspace;
+	GcmDeviceTypeEnum type;
 	GcmDevice *device = NULL;
 	gboolean ret;
+	gboolean virtual;
 	GError *error = NULL;
 	GcmClientPrivate *priv = client->priv;
 
@@ -648,17 +649,47 @@ gcm_client_add_unconnected_device (GcmClient *client, GKeyFile *keyfile, const g
 	title = g_key_file_get_string (keyfile, id, "title", NULL);
 	if (title == NULL)
 		goto out;
+	virtual = g_key_file_get_boolean (keyfile, id, "virtual", NULL);
 	type_text = g_key_file_get_string (keyfile, id, "type", NULL);
-	type = gcm_device_type_from_text (type_text);
+	type = gcm_device_type_enum_from_string (type_text);
+	if (type == GCM_DEVICE_TYPE_ENUM_UNKNOWN)
+		goto out;
+
+	/* get colorspace */
+	colorspace_text = g_key_file_get_string (keyfile, id, "colorspace", NULL);
+	if (colorspace_text == NULL) {
+		egg_warning ("legacy device %s, falling back to RGB", id);
+		colorspace = GCM_COLORSPACE_ENUM_RGB;
+	} else {
+		colorspace = gcm_colorspace_enum_from_string (colorspace_text);
+	}
+
+	/* create device or specified type */
+	if (virtual) {
+		device = gcm_device_virtual_new ();
+	} else if (type == GCM_DEVICE_TYPE_ENUM_DISPLAY) {
+		device = gcm_device_xrandr_new ();
+	} else if (type == GCM_DEVICE_TYPE_ENUM_PRINTER) {
+		device = gcm_device_cups_new ();
+	} else if (type == GCM_DEVICE_TYPE_ENUM_CAMERA) {
+		/* FIXME: use GPhoto? */
+		device = gcm_device_udev_new ();
+	} else if (type == GCM_DEVICE_TYPE_ENUM_SCANNER) {
+		device = gcm_device_sane_new ();
+	} else {
+		egg_warning ("device type internal error");
+		goto out;
+	}
 
 	/* create device */
-	device = gcm_device_new ();
 	g_object_set (device,
 		      "type", type,
 		      "id", id,
 		      "connected", FALSE,
 		      "title", title,
 		      "saved", TRUE,
+		      "virtual", virtual,
+		      "colorspace", colorspace,
 		      NULL);
 
 	/* load the device */
@@ -679,6 +710,7 @@ out:
 	if (device != NULL)
 		g_object_unref (device);
 	g_free (type_text);
+	g_free (colorspace_text);
 	g_free (title);
 }
 
@@ -709,8 +741,7 @@ gcm_client_add_saved (GcmClient *client, GError **error)
 	groups = g_key_file_get_groups (keyfile, NULL);
 	if (groups == NULL) {
 		ret = FALSE;
-		if (error != NULL)
-			*error = g_error_new (1, 0, "failed to get groups");
+		g_set_error_literal (error, 1, 0, "failed to get groups");
 		goto out;
 	}
 
@@ -741,22 +772,97 @@ gboolean
 gcm_client_add_connected (GcmClient *client, GError **error)
 {
 	gboolean ret;
+	GThread *thread;
 
 	g_return_val_if_fail (GCM_IS_CLIENT (client), FALSE);
 
-	/* usb */
-	ret = gcm_client_add_connected_devices_usb (client, error);
-	if (!ret)
-		goto out;
-
-	/* xorg */
+	/* XRandR */
 	ret = gcm_client_add_connected_devices_xrandr (client, error);
 	if (!ret)
 		goto out;
+
+	/* inform UI if we are loading devces still */
+	client->priv->loading_refcount = 2;
+	gcm_client_set_loading (client, TRUE);
+
+	/* UDEV */
+	if (client->priv->use_threads) {
+		thread = g_thread_create ((GThreadFunc) gcm_client_add_connected_devices_udev_thrd, client, FALSE, error);
+		if (thread == NULL)
+			goto out;
+	} else {
+		ret = gcm_client_add_connected_devices_udev (client, error);
+		if (!ret)
+			goto out;
+	}
+
+	/* CUPS */
+	if (client->priv->use_threads) {
+		thread = g_thread_create ((GThreadFunc) gcm_client_add_connected_devices_cups_thrd, client, FALSE, error);
+		if (thread == NULL)
+			goto out;
+	} else {
+		ret = gcm_client_add_connected_devices_cups (client, error);
+		if (!ret)
+			goto out;
+	}
 out:
 	return ret;
 }
 
+/**
+ * gcm_client_add_device:
+ **/
+gboolean
+gcm_client_add_device (GcmClient *client, GcmDevice *device, GError **error)
+{
+	gboolean ret = FALSE;
+	gchar *id = NULL;
+	gboolean virtual;
+	GcmDevice *device_tmp = NULL;
+
+	g_return_val_if_fail (GCM_IS_CLIENT (client), FALSE);
+	g_return_val_if_fail (GCM_IS_DEVICE (device), FALSE);
+
+	/* check removable */
+	g_object_get (device,
+		      "virtual", &virtual,
+		      "id", &id,
+		      NULL);
+	if (!virtual) {
+		g_set_error_literal (error, 1, 0, "cannot add non-virtual devices");
+		goto out;
+	}
+
+	/* look to see if device already exists */
+	device_tmp = gcm_client_get_device_by_id (client, id);
+	if (device_tmp != NULL) {
+		/* TRANSLATORS: error message */
+		g_set_error_literal (error, 1, 0, _("This device already exists"));
+		goto out;
+	}
+
+	/* add to the array */
+	g_ptr_array_add (client->priv->array, g_object_ref (device));
+
+	/* update status */
+	g_object_set (device,
+		      "saved", FALSE,
+		      NULL);
+
+	/* emit a signal */
+	egg_debug ("emit added: %s", id);
+	g_signal_emit (client, signals[SIGNAL_ADDED], 0, device);
+
+	/* all okay */
+	ret = TRUE;
+out:
+	g_free (id);
+	if (device_tmp != NULL)
+		g_object_unref (device_tmp);
+	return ret;
+}
+
 /**
  * gcm_client_delete_device:
  **/
@@ -779,16 +885,14 @@ gcm_client_delete_device (GcmClient *client, GcmDevice *device, GError **error)
 		      "id", &id,
 		      NULL);
 	if (connected) {
-		if (error != NULL)
-			*error = g_error_new (1, 0, "device is still connected");
+		g_set_error_literal (error, 1, 0, "device is still connected");
 		goto out;
 	}
 
 	/* try to remove from array */
 	ret = g_ptr_array_remove (client->priv->array, device);
 	if (!ret) {
-		if (error != NULL)
-			*error = g_error_new (1, 0, "not found in device array");
+		g_set_error_literal (error, 1, 0, "not found in device array");
 		goto out;
 	}
 
@@ -803,7 +907,7 @@ gcm_client_delete_device (GcmClient *client, GcmDevice *device, GError **error)
 		goto out;
 
 	/* remove from the config file */
-	g_key_file_remove_group (keyfile, id, error);
+	ret = g_key_file_remove_group (keyfile, id, error);
 	if (!ret)
 		goto out;
 
@@ -849,6 +953,12 @@ gcm_client_get_property (GObject *object, guint prop_id, GValue *value, GParamSp
 	case PROP_DISPLAY_NAME:
 		g_value_set_string (value, priv->display_name);
 		break;
+	case PROP_LOADING:
+		g_value_set_boolean (value, priv->loading);
+		break;
+	case PROP_USE_THREADS:
+		g_value_set_boolean (value, priv->use_threads);
+		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -869,12 +979,32 @@ gcm_client_set_property (GObject *object, guint prop_id, const GValue *value, GP
 		g_free (priv->display_name);
 		priv->display_name = g_strdup (g_value_get_string (value));
 		break;
+	case PROP_USE_THREADS:
+		priv->use_threads = g_value_get_boolean (value);
+		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
 	}
 }
 
+/**
+ * gcm_client_randr_event_cb:
+ **/
+static void
+gcm_client_randr_event_cb (GcmScreen *screen, GcmClient *client)
+{
+	GnomeRROutput **outputs;
+	guint i;
+
+	egg_debug ("screens may have changed");
+
+	/* replug devices */
+	outputs = gcm_screen_get_outputs (screen, NULL);
+	for (i=0; outputs[i] != NULL; i++)
+		gcm_client_xrandr_add (client, outputs[i]);
+}
+
 /**
  * gcm_client_class_init:
  **/
@@ -895,6 +1025,22 @@ gcm_client_class_init (GcmClientClass *klass)
 				     G_PARAM_READWRITE);
 	g_object_class_install_property (object_class, PROP_DISPLAY_NAME, pspec);
 
+	/**
+	 * GcmClient:loading:
+	 */
+	pspec = g_param_spec_boolean ("loading", NULL, NULL,
+				      TRUE,
+				      G_PARAM_READABLE);
+	g_object_class_install_property (object_class, PROP_LOADING, pspec);
+
+	/**
+	 * GcmClient:use-threads:
+	 */
+	pspec = g_param_spec_boolean ("use-threads", NULL, NULL,
+				      TRUE,
+				      G_PARAM_READWRITE);
+	g_object_class_install_property (object_class, PROP_USE_THREADS, pspec);
+
 	/**
 	 * GcmClient::added
 	 **/
@@ -924,26 +1070,27 @@ gcm_client_class_init (GcmClientClass *klass)
 static void
 gcm_client_init (GcmClient *client)
 {
-	GError *error = NULL;
 	const gchar *subsystems[] = {"usb", "video4linux", NULL};
 
 	client->priv = GCM_CLIENT_GET_PRIVATE (client);
 	client->priv->display_name = NULL;
+	client->priv->loading_refcount = 0;
+	client->priv->use_threads = FALSE;
 	client->priv->array = g_ptr_array_new_with_free_func ((GDestroyNotify) g_object_unref);
-	client->priv->edid = gcm_edid_new ();
-	client->priv->dmi = gcm_dmi_new ();
+	client->priv->screen = gcm_screen_new ();
+	g_signal_connect (client->priv->screen, "outputs-changed",
+			  G_CALLBACK (gcm_client_randr_event_cb), client);
 
 	/* use GUdev to find devices */
 	client->priv->gudev_client = g_udev_client_new (subsystems);
 	g_signal_connect (client->priv->gudev_client, "uevent",
 			  G_CALLBACK (gcm_client_uevent_cb), client);
 
-	/* get screen */
-	client->priv->rr_screen = gnome_rr_screen_new (gdk_screen_get_default (), (GnomeRRScreenChanged) gcm_client_randr_event_cb, client, &error);
-	if (client->priv->rr_screen == NULL) {
-		egg_error ("failed to get rr screen: %s", error->message);
-		g_error_free (error);
-	}
+	/* for CUPS */
+	httpInitialize();
+
+	/* should be okay for localhost */
+	client->priv->http = httpConnectEncrypt (cupsServer (), ippPort (), cupsEncryption ());
 }
 
 /**
@@ -958,9 +1105,8 @@ gcm_client_finalize (GObject *object)
 	g_free (priv->display_name);
 	g_ptr_array_unref (priv->array);
 	g_object_unref (priv->gudev_client);
-	g_object_unref (priv->edid);
-	g_object_unref (priv->dmi);
-	gnome_rr_screen_destroy (priv->rr_screen);
+	g_object_unref (priv->screen);
+	httpClose (priv->http);
 
 	G_OBJECT_CLASS (gcm_client_parent_class)->finalize (object);
 }
